//=============================================================================
// Project:    SLP - An Interpreter for Super Logic Programs / Static Semantics
// Module:     nints.cxx
// Purpose:    Set of interpretations for critical default negations.
// Last Change:11.03.2002
// Language:   C++
// Author:     Stefan Brass
// Email:      Stefan.Brass@informatik.uni-giessen.de, sbrass@sis.pitt.edu
// Address:    Universitaet Giessen, Arndtstr. 2, 35392 Giessen, Germany
// Copyright:  (c) 2002 by Stefan Brass
// Licence:    Distributed under the conditions of the GNU Public Licence.
//=============================================================================

//=============================================================================
// Interface:
//=============================================================================

#include "ver.h"
#include "out.h"
#include "check.h"
#include "nint.h"

#include "nints.h"

//=============================================================================
// Static Data Members (Class Attributes):
//=============================================================================

//-----------------------------------------------------------------------------
// Was the entire hash table set to NIL?
//-----------------------------------------------------------------------------

bool_t nints_c::Initialized = BOOL_FALSE;

//-----------------------------------------------------------------------------
// The hash table:
//-----------------------------------------------------------------------------

nints_c* nints_c::HashTab[NINTS_HASHSIZE];

//-----------------------------------------------------------------------------
// First element in the list of all default negation interpretations:
//-----------------------------------------------------------------------------

nints_c* nints_c::First = 0;

//-----------------------------------------------------------------------------
// Last element in the list of all default negation interpretations:
//-----------------------------------------------------------------------------

nints_c* nints_c::Last = 0;

//-----------------------------------------------------------------------------
// Deleted objects (can be reused):
//-----------------------------------------------------------------------------

nints_c* nints_c::Free = 0;

//=============================================================================
// Exported Static Member Functions (Class Methods):
//=============================================================================

//-----------------------------------------------------------------------------
// clear: Reinitialize to empty set (reuse memory).
//-----------------------------------------------------------------------------

void nints_c::clear(void)
{
	nints_t* hashtab_p;
	nints_t  elem, next_elem;
	int      i;

	// If the hash table is not initialized, make all entries NIL:
	if(!Initialized) {
		hashtab_p = HashTab;
		for(i = NINTS_HASHSIZE; --i >= 0; )
			*hashtab_p++ = 0;
		Initialized = BOOL_TRUE;
		CHECK(First == 0, "nints_c::clear: First not NIL");
		CHECK(Last  == 0, "nints_c::clear: Last not NIL");
		CHECK(Free  == 0, "nints_c::clear: Free not NIL");
		return;
	}

	// Clear used hash table elements, add to free list:
	for(elem = First; elem; elem = next_elem) {
		next_elem = elem->get_next();
		elem->free();
	}

	// Reset pointers to first and last element:
	First = 0;
	Last = 0;
}

//-----------------------------------------------------------------------------
// insert: Insert default negation interpretation if not already there.
//-----------------------------------------------------------------------------

void nints_c::insert(nint_t neg_int)
{
	nints_t* hash_p;
	nints_t  elem;

	// The module must be initialized:
	CHECK(Initialized, "nints_c::insert: Hash table not initialized");

	// Check parameters:
	CHECK(nint_c::valid(neg_int), "nints_c::insert: neg_int is invalid");

	// Check whether element is already contained in the set:
	hash_p = HashTab + nint_c::hash(neg_int, NINTS_HASHSIZE);
	for(elem = *hash_p; elem; elem = elem->get_hashlink())
		if(elem->get_neg_int() == neg_int)
			return;

	// Get free node, either from free list or from real constructor:
	if(Free) {
		elem = Free;
		Free = Free->get_next_free();
	}
	else
		elem = new nints_c;

	// Initialize the new hash chain element:
	elem->hash_enter(hash_p, neg_int);

	// Link new element to list of all set elements:
	if(First == 0) {
		CHECK(Last == 0, "nints_c::insert: Last must be NIL, too");
		First = elem;
		Last = elem;
	}
	else {
		CHECK(Last != 0, "nints_c::insert: Last must be not NIL, too");
		Last->set_next(elem);
		Last = elem;
	}
}

//-----------------------------------------------------------------------------
// print_all: Print all currently existing disjunctions.
//-----------------------------------------------------------------------------

void nints_c::print_all(void)
{
	nints_t elem;

	// The module must be initialized:
	CHECK(Initialized, "nints_c::print all: Hash table not initialized");

	// Check for empty set:
	if(First == 0) {
		out_c::no_min_mod();
		return;
	}

	// Start unordered list (if HTML output):
	out_c::list_begin();

	// Loop over all elements of the set (negation interpretations):
	for(elem = First; elem; elem = elem->get_next()) {
		out_c::list_item();
		nint_c::print(elem->get_neg_int());
	}

	// End unordered list (if HTML output):
	out_c::list_end();
}

//-----------------------------------------------------------------------------
// Debugging Support:
//-----------------------------------------------------------------------------

#ifdef	VER_DEBUG

bool_t	nints_c::valid(void) const
{
	// Check object pointer:
	CHECK(this != 0, "nints_c::valid: NIL object");

	// Magic number must be correct:
	if(Magic != NINTS_MAGIC)
		return(BOOL_FALSE);

	// The module must be initialized:
	CHECK(Initialized, "nints_c::valid: Hash table not initialized");

	// The default negation interpretation must be valid:
	CHECK(nint_c::valid(NegInt), "nints_c::valid: NegInt is invalid");

	// The corresponding hash table entry must be valid and set:
	CHECK(HashTabEntry >= HashTab && HashTabEntry < HashTab+NINTS_HASHSIZE,
		"nints_c::valid: Invalid HashTabEntry");
	CHECK(*HashTabEntry != 0, "nints_c::valid: Hash table entry is NIL");

	// The object cannot be on the free list (very partial check):
	CHECK(Free != this, "nints_c::valid: object on free list");

	// Ok:
	return(BOOL_TRUE);
}

#endif


